from fitsnap3.solvers.solver import Solver
"""Methods you may or must override in new solvers"""


class Template(Solver):

    def __init__(self, name):
        super().__init__(name)

    # Solver must override perform_fit method
    def perform_fit(self):
        """"""
        pass



from fitsnap3.io.sections.sections import Section


class Solver(Section):

    def __init__(self, name, config, args):
        super().__init__(name, config, args)
        self.solver = self.get_value("SOLVER", "solver", "SVD")
        self.normalweight = self.get_value("SOLVER", "normalweight", "-12", "float")
        self.normratio = self.get_value("SOLVER", "normratio", "0.5", "float")
        self.compute_dbvb = self.get_value("SOLVER", "compute_dbvb", "0", "bool")
        self.compute_testerrs = self.get_value("SOLVER", "compute_testerrs", "0", "bool")
        self.multinode_testing = self.get_value("SOLVER", "multinode_testing", "0", "bool")
        self.delete()



from fitsnap3.io.sections.sections import Section


class Model(Section):

    def __init__(self, name, config, args):
        super().__init__(name, config, args)
        self.chemflag = self.get_value("MODEL", "chemflag", "0", "bool")
        self.wselfallflag = self.get_value("MODEL", "wselfallflag", "0", "bool")
        self.bzeroflag = self.get_value("MODEL", "bzeroflag", "0", "bool")
        self.quadraticflag = self.get_value("MODEL", "quadraticflag", "0", "bool")
        # self.solver = self.get_value("MODEL", "solver", "SVD")
        # self.normalweight = self.get_value("MODEL", "normalweight", "-12", "float")
        # self.normratio = self.get_value("MODEL", "normratio", "0.5", "float")
        # self.compute_dbvb = self.get_value("MODEL", "compute_dbvb", "0", "bool")
        # self.compute_testerrs = self.get_value("MODEL", "compute_testerrs", "0", "bool")
        # self.delete()



import ctypes
from fitsnap3.calculators.calculator import Calculator
from fitsnap3.parallel_tools import pt
from fitsnap3.io.input import config
import numpy as np

# TODO: Ask about onehot_fraction for energy


class LammpsSnap(Calculator):

    def __init__(self, name):
        super().__init__(name)
        self._data = {}
        self._i = 0
        self._lmp = None

    def create_a(self):
        super().create_a()

    def process_configs(self, data, i):
        self._data = data
        self._i = i
        self._initialize_lammps()
        self._prepare_lammps()
        self._run_lammps()
        self._collect_lammps()
        self._lmp = pt.close_lammps()

    def _initialize_lammps(self):
        self._lmp = pt.initialize_lammps(config.args.lammpslog, config.args.printlammps)

    def _prepare_lammps(self):
        self._lmp.command("clear")
        self._lmp.command("units " + config.sections["REFERENCE"].units)
        self._lmp.command("atom_style " + config.sections["REFERENCE"].atom_style)

        lmp_setup = _extract_commands("""
                atom_modify map array sort 0 2.0
                box tilt large""")
        for line in lmp_setup:
            self._lmp.command(line)

        self._set_box()

        self._create_atoms()

        if config.sections["REFERENCE"].atom_style == "spin":
            self._create_spins()
        if config.sections["REFERENCE"].atom_style == "charge":
            self._create_charge()

        self._set_variables(**_lammps_variables(config.sections["BISPECTRUM"].__dict__))

        for line in config.sections["REFERENCE"].lmp_pairdecl:
            self._lmp.command(line.lower())

        self._set_computes()

        self._lmp.command("mass * 1.0e-20")
        self._lmp.command("neighbor 1.0e-20 nsq")
        self._lmp.command("neigh_modify one 10000")

    def _set_box(self):
        self._lmp.command("boundary p p p")
        ((ax, bx, cx),
         (ay, by, cy),
         (az, bz, cz)) = self._data["Lattice"]

        assert all(abs(c) < 1e-10 for c in (ay, az, bz)), \
            "Cell not normalized for lammps!"
        region_command = \
            f"region pybox prism 0 {ax:20.20g} 0 {by:20.20g} 0 {cz:20.20g} {bx:20.20g} {cx:20.20g} {cy:20.20g}"
        self._lmp.command(region_command)
        self._lmp.command(f"create_box {config.sections['BISPECTRUM'].numtypes} pybox")

    def _create_atoms(self):
        for i, (a_t, (a_x, a_y, a_z)) in enumerate(zip(self._data["AtomTypes"], self._data["Positions"])):
            a_t = config.sections["BISPECTRUM"].type_mapping[a_t]
            self._lmp.command(f"create_atoms {a_t} single {a_x:20.20g} {a_y:20.20g} {a_z:20.20g} remap yes")
        n_atoms = int(self._lmp.get_natoms())
        assert i + 1 == n_atoms, f"Atom counts don't match when creating atoms: {i + 1}, {n_atoms}"

    def _create_spins(self):
        for i, (s_mag, s_x, s_y, s_z) in enumerate(self._data["Spins"]):
            self._lmp.command(f"set atom {i + 1} spin {s_mag:20.20g} {s_x:20.20g} {s_y:20.20g} {s_z:20.20g} ")
        n_atoms = int(self._lmp.get_natoms())
        assert i + 1 == n_atoms, f"Atom counts don't match when assigning spins: {i + 1}, {n_atoms}"

    def _create_charge(self):
        for i, q in enumerate(self._data["Charges"]):
            self._lmp.command(f"set atom {i + 1} charge {q:20.20g} ")
        n_atoms = int(self._lmp.get_natoms())
        assert i + 1 == n_atoms, f"Atom counts don't match when assigning charge: {i + 1}, {n_atoms}"

    def _set_variables(self, **lmp_variable_args):
        for k, v in lmp_variable_args.items():
            self._lmp.command(f"variable {k} equal {v}")

    def _set_computes(self):
        numtypes = config.sections['BISPECTRUM'].numtypes
        radelem = " ".join([f"${{radelem{i}}}" for i in range(1, numtypes + 1)])
        wj = " ".join([f"${{wj{i}}}" for i in range(1, numtypes + 1)])

        kw_options = {
            k: config.sections["CALCULATOR"].__dict__[v]
            for k, v in
            {
                "rmin0": "rmin0",
                "bzeroflag": "bzeroflag",
                "quadraticflag": "quadraticflag",
                "switchflag": "switchflag",
                "chem": "chemflag",
                "bnormflag": "bnormflag",
                "wselfallflag": "wselfallflag",
            }.items()
            if v in config.sections["CALCULATOR"].__dict__
        }
        if kw_options["chem"] == 0:
            kw_options.pop("chem")
        kw_options["rmin0"] = config.sections["BISPECTRUM"].rmin0
        kw_substrings = [f"{k} {v}" for k, v in kw_options.items()]
        kwargs = " ".join(kw_substrings)

        # everything is handled by LAMMPS compute snap

        base_snap = "compute snap all snap ${rcutfac} ${rfac0} ${twojmax}"
        command = f"{base_snap} {radelem} {wj} {kwargs}"
        self._lmp.command(command)

    def _run_lammps(self):
        self._lmp.command("run 0")

    def _collect_lammps(self):

        num_atoms = self._data["NumAtoms"]
        num_types = config.sections['BISPECTRUM'].numtypes
        n_coeff = config.sections['BISPECTRUM'].ncoeff
        energy = self._data["Energy"]

        lmp_atom_ids = self._lmp.numpy.extract_atom_iarray("id", num_atoms).ravel()
        assert np.all(lmp_atom_ids == 1 + np.arange(num_atoms)), "LAMMPS seems to have lost atoms"

        # Extract positions
        lmp_pos = self._lmp.numpy.extract_atom_darray(name="x", nelem=num_atoms, dim=3)
        # Extract types
        lmp_types = self._lmp.numpy.extract_atom_iarray(name="type", nelem=num_atoms).ravel()
        lmp_volume = self._lmp.get_thermo("vol")

        # Extract SNAP data, including reference potential data

        nrows_energy = 1
        ndim_force = 3
        nrows_force = ndim_force * num_atoms
        ndim_virial = 6
        nrows_virial = ndim_virial
        nrows_snap = nrows_energy + nrows_force + nrows_virial
        ncols_bispectrum = num_types * n_coeff
        ncols_reference = 1
        ncols_snap = ncols_bispectrum + ncols_reference
        index = pt.shared_arrays['a'].indices[self._i]

        lmp_snap = _extract_compute_np(self._lmp, "snap", 0, 2, (nrows_snap, ncols_snap))
        if (np.isinf(lmp_snap)).any() or (np.isnan(lmp_snap)).any():
            raise ValueError('Nan in computed data of file {} in group {}'.format(self._data["File"],
                                                                                  self._data["Group"]))

        irow = 0
        icolref = ncols_bispectrum
        if config.sections["CALCULATOR"].energy:
            b_sum_temp = lmp_snap[irow, :ncols_bispectrum] / num_atoms
            if not config.sections["CALCULATOR"].bzeroflag:
                b_sum_temp.shape = (num_types, n_coeff)
                onehot_atoms = np.ones((num_types, 1))
                b_sum_temp = np.concatenate((onehot_atoms, b_sum_temp), axis=1)
                b_sum_temp.shape = (num_types * n_coeff + num_types)
            # fix onehot_atoms to be true fraction not just 1
            # onehot_atoms = np.stack([np.eye(num_types)[x - 1, :].sum(axis=0) for x in lmp_types])
            # onehot_fraction = onehot_atoms / onehot_atoms.sum(axis=1)[:, np.newaxis
            pt.shared_arrays['a'].array[index] = b_sum_temp
            ref_energy = lmp_snap[irow, icolref]
            pt.shared_arrays['b'].array[index] = (energy - ref_energy) / num_atoms
            pt.shared_arrays['w'].array[index] = self._data["eweight"]
            irow += nrows_energy
            index += 1

        if config.sections["CALCULATOR"].force:
            db_atom_temp = lmp_snap[irow:irow + nrows_force, :ncols_bispectrum]
            db_atom_temp.shape = (num_atoms * ndim_force, n_coeff * num_types)
            if not config.sections["CALCULATOR"].bzeroflag:
                onehot_atoms = np.zeros((np.shape(db_atom_temp)[0],))
                db_atom_temp = np.concatenate([onehot_atoms[:, np.newaxis], db_atom_temp], axis=1)
            pt.shared_arrays['a'].array[index:index+num_atoms * ndim_force] = db_atom_temp
            ref_forces = lmp_snap[irow:irow + nrows_force, icolref]
            pt.shared_arrays['b'].array[index:index+num_atoms * ndim_force] = \
                self._data["Forces"].ravel() - ref_forces
            pt.shared_arrays['w'].array[index:index+num_atoms * ndim_force] = \
                self._data["fweight"]
            irow += nrows_force
            index += num_atoms * ndim_force

        if config.sections["CALCULATOR"].stress:
            vb_sum_temp = 1.6021765e6*lmp_snap[irow:irow + nrows_virial, :ncols_bispectrum] / lmp_volume
            vb_sum_temp.shape = (ndim_virial, n_coeff * num_types)
            if not config.sections["CALCULATOR"].bzeroflag:
                onehot_atoms = np.zeros((np.shape(vb_sum_temp)[0],))
                vb_sum_temp = np.concatenate([onehot_atoms[:, np.newaxis], vb_sum_temp], axis=1)
            pt.shared_arrays['a'].array[index:index+ndim_virial] = vb_sum_temp
            ref_stress = lmp_snap[irow:irow + nrows_virial, icolref]
            pt.shared_arrays['b'].array[index:index+ndim_virial] = \
                self._data["Stress"][[0, 1, 2, 1, 0, 0], [0, 1, 2, 2, 2, 1]].ravel() - ref_stress
            pt.shared_arrays['w'].array[index:index+ndim_virial] = \
                self._data["vweight"]
            index += ndim_virial


def _lammps_variables(bispec_options):
    d = {k: bispec_options[k] for k in
         ["rcutfac",
          "rfac0",
          "rmin0",
          "twojmax"]}
    d.update(
        {
            (k + str(i + 1)): bispec_options[k][i]
            # "zblz", "wj", "radelem"
            for k in ["wj", "radelem"]
            for i, v in enumerate(bispec_options[k])
        })
    return d


def _extract_compute_np(lmp, name, compute_style, result_type, array_shape):
    """
    Convert a lammps compute to a numpy array.
    Assumes the compute stores floating point numbers.
    Note that the result is a view into the original memory.
    If the result type is 0 (scalar) then conversion to numpy is
    skipped and a python float is returned.

    From LAMMPS/src/library.cpp:
    style = 0 for global data, 1 for per-atom data, 2 for local data
    type = 0 for scalar, 1 for vector, 2 for array

    """
    ptr = lmp.extract_compute(name, compute_style, result_type)
    if result_type == 0:
        # No casting needed, lammps.py already works
        return ptr
    if result_type == 2:
        ptr = ptr.contents
    total_size = np.prod(array_shape)
    buffer_ptr = ctypes.cast(ptr, ctypes.POINTER(ctypes.c_double * total_size))
    array_np = np.frombuffer(buffer_ptr.contents, dtype=float)
    array_np.shape = array_shape
    return array_np


def _extract_commands(string):
    return [x for x in string.splitlines() if x.strip() != '']



from fitsnap3.io.outputs.outputs import Output, optional_write
from fitsnap3.parallel_tools import pt
from datetime import datetime
from fitsnap3.io.input import config


class Original(Output):

    def __init__(self, name):
        super().__init__(name)

    def output(self, coeffs, errors):
        new_coeffs = pt.combine_coeffs(coeffs)
        if new_coeffs is not None:
            coeffs = new_coeffs
        self.write(coeffs, errors)

    @pt.rank_zero
    def write(self, coeffs, errors):
        with optional_write(config.sections["OUTFILE"].potential_name and
                            config.sections["OUTFILE"].potential_name + '.snapcoeff', 'wt') as file:
            file.write(_to_coeff_string(coeffs))
        with optional_write(config.sections["OUTFILE"].potential_name and
                            config.sections["OUTFILE"].potential_name + '.snapparam', 'wt') as file:
            file.write(_to_param_string())
        with optional_write(config.sections["OUTFILE"].metric_file, 'wt') as file:
            errors.to_csv(file)


def _to_param_string():
    if config.sections["CALCULATOR"].chemflag != 0:
        chemflag_int = 1
    else:
        chemflag_int = 0
    return f"""
    # required
    rcutfac {config.sections["BISPECTRUM"].rcutfac}
    twojmax {config.sections["BISPECTRUM"].twojmax}

    #  optional
    rfac0 {config.sections["BISPECTRUM"].rfac0}
    rmin0 {config.sections["BISPECTRUM"].rmin0}
    bzeroflag {config.sections["CALCULATOR"].bzeroflag}
    quadraticflag {config.sections["CALCULATOR"].quadraticflag}
    wselfallflag {config.sections["CALCULATOR"].wselfallflag}
    chemflag {chemflag_int}
    """


def _to_coeff_string(coeffs):
    """
    Convert a set of coefficients along with bispec options into a .snapparam file
    """
    # Includes the offset term, which was not in blist
    coeffs = coeffs.reshape((config.sections["BISPECTRUM"].numtypes, -1))
    coeff_names = [[0]]+config.sections["BISPECTRUM"].blist
    type_names = config.sections["BISPECTRUM"].types
    out = f"# fitsnap fit generated on {datetime.now()}\n\n"
    out += "{} {}\n".format(len(type_names), len(coeff_names))
    for elname, rjval, wjval, column in zip(type_names,
                                            config.sections["BISPECTRUM"].radelem,
                                            config.sections["BISPECTRUM"].wj,
                                            coeffs):
        out += "{} {} {}\n".format(elname, rjval, wjval)
        out += "\n".join(f" {bval:<30.18} #  B{bname} " for bval, bname in zip(column, coeff_names))
        out += "\n"
    out += "\n# End of potential"
    return out



from fitsnap3.solvers.solver import Solver
from fitsnap3.parallel_tools import pt
from fitsnap3.io.input import config
import numpy as np
try:
    import tensorflow as tf


    class TensorflowSVD(Solver):

        def __init__(self, name):
            super().__init__(name)

        @pt.single_timeit
        def perform_fit(self, testing):
            w = pt.shared_arrays['w'].array
            aw, bw = w[:, np.newaxis] * pt.shared_arrays['a'].array, w * pt.shared_arrays['b'].array
            bw = bw.reshape((len(bw), 1))
            self.fit = tf.linalg.lstsq(aw, bw)
            if config.sections["CALCULATOR"].bzeroflag:
                self._offset()

except ModuleNotFoundError:

    class TensorflowSVD(Solver):

        def __init__(self, name):
            super().__init__(name)
            raise ModuleNotFoundError("No module named 'tensorflow'")



from fitsnap3.io.sections.sections import Section
from fitsnap3.parallel_tools import pt


class Calculator(Section):

    def __init__(self, name, config, args):
        super().__init__(name, config, args)
        self.calculator = self.get_value("CALCULATOR", "calculator", "LAMMPSSNAP")
        self.energy = self.get_value("CALCULATOR", "energy", "True", "bool")
        pt.add_calculator_option("energy", self.energy)
        self.force = self.get_value("CALCULATOR", "force", "True", "bool")
        pt.add_calculator_option("force", self.force)
        self.stress = self.get_value("CALCULATOR", "stress", "True", "bool")
        pt.add_calculator_option("stress", self.stress)
        self.chemflag = self.get_value("CALCULATOR", "chemflag", "0", "bool")
        self.bnormflag = self.get_value("CALCULATOR", "bnormflag", "0", "bool")
        self.wselfallflag = self.get_value("CALCULATOR", "wselfallflag", "0", "bool")
        self.bzeroflag = self.get_value("CALCULATOR", "bzeroflag", "0", "bool")
        self.quadraticflag = self.get_value("CALCULATOR", "quadraticflag", "0", "bool")
        self.delete()



from fitsnap3.solvers.solver import Solver
from fitsnap3.solvers.template_solver import Template
from fitsnap3.solvers.svd import SVD
from fitsnap3.solvers.tensorflowsvd import TensorflowSVD


def solver(solver_name):
    """Solver Factory"""
    instance = search(solver_name)
    instance.__init__(solver_name)
    return instance


def search(solver_name):
    instance = None
    for cls in Solver.__subclasses__():
        if cls.__name__.lower() == solver_name.lower():
            instance = Solver.__new__(cls)

    if instance is None:
        raise IndexError("{} was not found in fitsnap solvers".format(solver_name))
    else:
        return instance



from os import path
from fitsnap3.io.sections.sections import Section


class Outfile(Section):

    def __init__(self, name, config, args):
        super().__init__(name, config, args)
        self._check_relative()
        self._outfile()
        self.output_style = self.get_value("OUTFILE", "output_style", "ORIGINAL")
        self.delete()

    def _outfile(self):
        self.config_file = self.get_value("OUTFILE", "configs", "fitsnap_configs.pkl.gz")
        self._check_path(self.config_file)
        self.metric_file = self.get_value("OUTFILE", "metrics", "fitsnap_metrics.csv")
        self._check_path(self.metric_file)
        self.potential_name = self.get_value("OUTFILE", "potential", "fitsnap_potential")
        self._check_path(self.potential_name)
        return

    def _check_relative(self):
        if self._args.relative:
            self.base_path, _ = path.split(self._args.infile)
        else:
            self.base_path = None

    def _check_path(self, name):
        if self.base_path is not None:
            name = path.join(self.base_path, name)
        else:
            name = name
        if self._args.overwrite is None:
            return name
        names = [name, name + '.snapparam', name + '.snapcoeff']
        for element in names:
            if self._args.overwrite and path.exists(element):
                raise FileExistsError(f"File {element} already exists.")


from fitsnap3.parallel_tools import pt, double_size
from fitsnap3.io.input import config
from fitsnap3.io.output import output


class Calculator:

    def __init__(self, name):
        self.name = name
        self.number_of_atoms = None
        self.number_of_files_per_node = None

    def create_a(self):
        # TODO : Any extra config pulls should be done before this
        pt.sub_barrier()
        self.number_of_atoms = pt.shared_arrays["number_of_atoms"].array.sum()
        self.number_of_files_per_node = len(pt.shared_arrays["number_of_atoms"].array)
        num_types = config.sections["BISPECTRUM"].numtypes

        a_len = 0
        if config.sections["CALCULATOR"].energy:
            a_len += self.number_of_files_per_node
        if config.sections["CALCULATOR"].force:
            a_len += 3 * self.number_of_atoms
        if config.sections["CALCULATOR"].stress:
            a_len += self.number_of_files_per_node * 6

        a_width = config.sections["BISPECTRUM"].ncoeff * num_types
        if not config.sections["CALCULATOR"].bzeroflag:
            a_width += num_types

        # TODO: Pick a method to get RAM accurately (pt.get_ram() seems to get RAM wrong on Blake)
        a_size = a_len * a_width * double_size
        # output.screen("'a' takes up ", 100 * a_size / pt.get_ram(), "% of the total memory")
        output.screen("'a' takes up ", 100 * a_size / config.sections["MEMORY"].memory,
                      "% of the total memory", config.sections["MEMORY"].memory*1e-9, "GB of RAM")
        if a_size / pt.get_ram() > 0.5 and not config.sections["MEMORY"].override:
            raise MemoryError("The 'a' matrix is larger than 50% of your RAM. \n Aborting...!")
        elif a_size / pt.get_ram() > 0.5 and config.sections["MEMORY"].override:
            output.screen("Warning: I hope you know what your are doing!")

        pt.create_shared_array('a', a_len, a_width)
        pt.create_shared_array('b', a_len)
        pt.create_shared_array('w', a_len)
        pt.slice_array('a', num_types=num_types)

    def process_configs(self, data, i):
        pass



from fitsnap3.io.sections.sections import Section
from os import sysconf
from subprocess import check_output


class Memory(Section):

    def __init__(self, name, config, args):
        super().__init__(name, config, args)
        self._check_memory()
        self.memory = self.get_value("MEMORY", "memory", "{}".format(self.mem_bytes), "int")
        self.override = self.get_value("MEMORY", "override", "False", interpreter="bool")
        self.delete()

    def _check_memory(self):
        try:
            self.mem_bytes = sysconf('SC_PAGE_SIZE') * sysconf('SC_PHYS_PAGES')
        except ValueError:
            self.mem_bytes = int(check_output(['sysctl', '-n', 'hw.memsize']).strip())



# <!----------------BEGIN-HEADER------------------------------------>
# ## FitSNAP3
# A Python Package For Training SNAP Interatomic Potentials for use in the LAMMPS molecular dynamics package
#
# _Copyright (2016) Sandia Corporation.
# Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
# the U.S. Government retains certain rights in this software.
# This software is distributed under the GNU General Public License_
# ##
#
# #### Original author:
#     Aidan P. Thompson, athomps (at) sandia (dot) gov (Sandia National Labs)
#     http://www.cs.sandia.gov/~athomps
#
# #### Key contributors (alphabetical):
#     Mary Alice Cusentino (Sandia National Labs)
#     Nicholas Lubbers (Los Alamos National Lab)
#     Charles Sievers (Sandia National Labs)
#     Adam Stephens (Sandia National Labs)
#     Mitchell Wood (Sandia National Labs)
#
# #### Additional authors (alphabetical):
#     Elizabeth Decolvenaere (D. E. Shaw Research)
#     Stan Moore (Sandia National Labs)
#     Steve Plimpton (Sandia National Labs)
#     Gary Saavedra (Sandia National Labs)
#     Peter Schultz (Sandia National Labs)
#     Laura Swiler (Sandia National Labs)
#
# <!-----------------END-HEADER------------------------------------->

try:
    import mpi4py as mpi4py
    from fitsnap3.parallel_tools import pt

except ModuleNotFoundError:
    from fitsnap3.parallel_tools import pt

except Exception as e:
    print("Trouble importing mpi4py package, exiting...")
    raise e

pt.single_print("")
pt.single_print("    ______ _  __  _____  _   __ ___     ____  ")
pt.single_print("   / ____/(_)/ /_/ ___/ / | / //   |   / __ \ ")
pt.single_print("  / /_   / // __/\__ \ /  |/ // /| |  / /_/ /")
pt.single_print(" / __/  / // /_ ___/ // /|  // ___ | / ____/ ")
pt.single_print("/_/    /_/ \__//____//_/ |_//_/  |_|/_/      ")
pt.single_print("")
pt.single_print("-----------")

try:
    pt.single_print("Reading input...")
    pt.all_barrier()
    from fitsnap3.io.input import config
    pt.single_print("Finished reading input")
    from fitsnap3.io.output import output
except Exception as e:
    pt.single_print("Trouble reading input, exiting...")
    pt.exception(e)

try:
    output.screen("mpi4py version: ", mpi4py.__version__)

except NameError:
    print("No mpi4py detected, using fitsnap stubs...")

try:
    import numpy as np
    output.screen("numpy version: ", np.__version__)
except Exception as e:
    output.screen("Trouble importing numpy package, exiting...")
    output.exception(e)

try:
    import scipy as sp
    output.screen("scipy version: ", sp.__version__)
except Exception as e:
    output.screen("Trouble importing scipy package, exiting...")
    output.exception(e)

try:
    import pandas as pd
    output.screen("pandas version: ", pd.__version__)
except Exception as e:
    output.screen("Trouble importing pandas package, exiting...")
    output.exception(e)

# try:
#     import lammps
#     print("LAMMPS version: ",lammps.version())
# except Exception as e:
#     print("Trouble importing LAMMPS library, exiting...")
#     raise e
output.screen("-----------")



from fitsnap3.io.sections.sections import Section


class Scraper(Section):

    def __init__(self, name, config, args):
        super().__init__(name, config, args)
        self.scraper = self.get_value("SCRAPER", "scraper", "JSON")
        self.delete()



# <!----------------BEGIN-HEADER------------------------------------>
# ## FitSNAP3
# A Python Package For Training SNAP Interatomic Potentials for use in the LAMMPS molecular dynamics package
#
# _Copyright (2016) Sandia Corporation.
# Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
# the U.S. Government retains certain rights in this software.
# This software is distributed under the GNU General Public License_
# ##
#
# #### Original author:
#     Aidan P. Thompson, athomps (at) sandia (dot) gov (Sandia National Labs)
#     http://www.cs.sandia.gov/~athomps
#
# #### Key contributors (alphabetical):
#     Mary Alice Cusentino (Sandia National Labs)
#     Nicholas Lubbers (Los Alamos National Lab)
#     Charles Sievers (Sandia Nation Labs)
#     Adam Stephens (Sandia National Labs)
#     Mitchell Wood (Sandia National Labs)
#
# #### Additional authors (alphabetical):
#     Elizabeth Decolvenaere (D. E. Shaw Research)
#     Stan Moore (Sandia National Labs)
#     Steve Plimpton (Sandia National Labs)
#     Gary Saavedra (Sandia National Labs)
#     Peter Schultz (Sandia National Labs)
#     Laura Swiler (Sandia National Labs)
#
# <!-----------------END-HEADER------------------------------------->

from fitsnap3.fitsnap import FitSnap
from fitsnap3.parallel_tools import pt
from fitsnap3.io.output import output
# TODO : Make unit conversion object


@pt.single_timeit
def main():
    try:
        snap = FitSnap()
        snap.scrape_configs()
        snap.process_configs()
        pt.all_barrier()
        snap.perform_fit()
        snap.write_output()
    except Exception as e:
        output.exception(e)


if __name__ == "__main__":
    main()




from fitsnap3.io.sections.sections import Section


class Eshift(Section):

    def __init__(self, name, config, args):
        self._on = "False"
        super().__init__(name, config, args)
        types = []
        for i in range(self.get_value("BISPECTRUM", "numTypes", "1", "int")):
            types.append(self.get_value("BISPECTRUM", "type{}".format(i + 1), "H"))
        self.eshift = {}
        for atom_type in types:
            self.eshift[atom_type] = self.get_value("ESHIFT", "{}".format(atom_type), "0.0", "float")
        self.delete()



from fitsnap3.io.sections.sections import Section
from fitsnap3.io.sections.bispectrum import Bispectrum
from fitsnap3.io.sections.calculator import Calculator
from fitsnap3.io.sections.template import Default
from fitsnap3.io.sections.eshift import Eshift
from fitsnap3.io.sections.memory import Memory
from fitsnap3.io.sections.model import Model
from fitsnap3.io.sections.outfile import Outfile
from fitsnap3.io.sections.path import Path
from fitsnap3.io.sections.reference import Reference
from fitsnap3.io.sections.scraper import Scraper
from fitsnap3.io.sections.solver import Solver
from fitsnap3.io.sections.groups import Groups


def new_section(section, config, args):
    """Section Factory"""
    instance = search(section)
    instance.__init__(section, config, args)
    return instance


def search(section):
    instance = None
    for cls in Section.__subclasses__():
        if cls.__name__.lower() == section.lower():
            instance = Section.__new__(cls)

    if instance is None:
        raise IndexError("{} was not found in fitsnap sections".format(section))
    else:
        return instance



from fitsnap3.io.outputs.outputs import Output
"""Methods you may or must override in new output"""


class Template(Output):

    def __init__(self, name):
        super().__init__(name)

    # You must override this method
    def output(self):
        pass



from fitsnap3.io.outputs.output_factory import output_factory
from fitsnap3.io.input import config

if __name__ == "fitsnap3.io.output":
    output = output_factory(config.sections["OUTFILE"].output_style)



from fitsnap3.parallel_tools import pt, output
from distutils.util import strtobool


class Section:
    parameters = []

    def __init__(self, name, config, args):
        self.name = name
        self._config = config
        self._args = args
        try:
            on = self._on
        except AttributeError:
            on = "True"
        self._on = self.get_value(self.name.upper(), "on", on, "bool")
        if self._on is False:
            self.delete()
            return

    def delete(self):
        del self._config
        del self._args

    def print_name(self):
        output.screen(self.name)

    def get_value(self, section, key, fallback, interpreter="str"):
        if self._args == "verbose" and section.lower() == self.name.lower():
            Section.add_parameter(section, key, fallback, interpreter)
        if interpreter == "str" or interpreter == "string":
            convert = str
        elif interpreter == "bool":
            convert = strtobool
        elif interpreter == "float":
            convert = float
        elif interpreter == "int" or interpreter == "integer":
            convert = int
        else:
            raise ValueError("{} is not an implemented interpreter.")

        if section not in self._config:
            return convert(fallback)
        else:
            return convert(self._config.get(section, key, fallback=fallback))

    def get_section(self, section):
        if section not in self._config:
            return None
        return self._config.items(section)

    @classmethod
    def add_parameter(cls, section, key, fallback, interpreter):
        cls.parameters.append([section, key, fallback, interpreter])



from fitsnap3.scrapers.scrape import Scraper
"""Methods you may or must override in new scrapers"""


class Template(Scraper):

    def __init__(self, name):
        super().__init__(name)

    # Scraper may override scrape_groups method
    def scrape_groups(self):
        """Need self.files and self.group_table"""
        pass

    # Scraper must override scrape_configs method
    def scrape_configs(self):
        """Generate and send (mutable) data to send to fitsnap"""
        pass




from fitsnap3.scrapers.scrape import Scraper
from fitsnap3.scrapers.template_scraper import Template
from fitsnap3.scrapers.json_scraper import Json
# from fitsnap3.scrapers.xyz_scraper import XYZ


def scraper(scraper_name):
    """Section Factory"""
    instance = search(scraper_name)
    instance.__init__(scraper_name)
    return instance


def search(scraper_name):
    instance = None
    for cls in Scraper.__subclasses__():
        if cls.__name__.lower() == scraper_name.lower():
            instance = Scraper.__new__(cls)

    if instance is None:
        raise IndexError("{} was not found in fitsnap scrapers".format(scraper_name))
    else:
        return instance



from fitsnap3.io.outputs.outputs import Output
from fitsnap3.io.outputs.template_output import Template
from fitsnap3.io.outputs.original import Original


def output_factory(output_name):
    """Output Factory"""
    instance = search(output_name)
    instance.__init__(output_name)
    return instance


def search(output_name):
    instance = None
    for cls in Output.__subclasses__():
        if cls.__name__.lower() == output_name.lower():
            instance = Output.__new__(cls)

    if instance is None:
        raise IndexError("{} was not found in fitsnap outputs".format(output_name))
    else:
        return instance



from fitsnap3.io.sections.sections import Section
from itertools import combinations_with_replacement
import numpy as np


class Bispectrum(Section):

    def __init__(self, name, config, args):
        super().__init__(name, config, args)

        self.numtypes = self.get_value("BISPECTRUM", "numTypes", "1", "int")
        self.twojmax = self.get_value("BISPECTRUM", "twojmax", "6", "int")
        self.rcutfac = self.get_value("BISPECTRUM", "rcutfac", "4.67637", "float")
        self.rfac0 = self.get_value("BISPECTRUM", "rfac0", "0.99363", "float")
        self.rmin0 = self.get_value("BISPECTRUM", "rmin0", "0.0", "float")
        self.wj = []
        for i in range(self.numtypes):
            self.wj.append(self.get_value("BISPECTRUM", "wj{}".format(i + 1), "1.0", "float"))
        self.radelem = []
        for i in range(self.numtypes):
            self.radelem.append(self.get_value("BISPECTRUM", "radelem{}".format(i + 1), "0.5", "float"))
        self.types = []
        for i in range(self.numtypes):
            self.types.append(self.get_value("BISPECTRUM", "type{}".format(i + 1), "H"))
        self.type_mapping = {}
        for i, atom_type in enumerate(self.types):
            self.type_mapping[atom_type] = i+1

        self.boltz = self.get_value("BISPECTRUM", "BOLTZ", "1000", "float")
        self._generate_b_list()
        self.delete()

    def _generate_b_list(self):
        self.blist = []
        i = 0
        for j1 in range(self.twojmax + 1):
            for j2 in range(j1 + 1):
                for j in range(abs(j1 - j2), min(self.twojmax, j1 + j2) + 1, 2):
                    if j >= j1:
                        i += 1
                        self.blist.append([i, j1, j2, j])
        if self.get_value("CALCULATOR", "chemflag", "0", "bool"):
            self.blist *= self.numtypes ** 3
            self.blist = np.array(self.blist).tolist()
        if self.get_value("CALCULATOR", "quadraticflag", "0", "bool"):
            # Note, combinations_with_replacement precisely generates the upper-diagonal entries we want
            self.blist += [[i, a, b] for i, (a, b) in
                           enumerate(combinations_with_replacement(self.blist, r=2), start=len(self.blist))]
        self.ncoeff = len(self.blist)



# <!----------------BEGIN-HEADER------------------------------------>
# ## FitSNAP3
# A Python Package For Training SNAP Interatomic Potentials for use in the LAMMPS molecular dynamics package
#
# _Copyright (2016) Sandia Corporation.
# Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
# the U.S. Government retains certain rights in this software.
# This software is distributed under the GNU General Public License_
# ##
#
# #### Original author:
#     Aidan P. Thompson, athomps (at) sandia (dot) gov (Sandia National Labs)
#     http://www.cs.sandia.gov/~athomps
#
# #### Key contributors (alphabetical):
#     Mary Alice Cusentino (Sandia National Labs)
#     Nicholas Lubbers (Los Alamos National Lab)
#     Maybe me ¯\_(ツ)_/¯
#     Adam Stephens (Sandia National Labs)
#     Mitchell Wood (Sandia National Labs)
#
# #### Additional authors (alphabetical):
#     Elizabeth Decolvenaere (D. E. Shaw Research)
#     Stan Moore (Sandia National Labs)
#     Steve Plimpton (Sandia National Labs)
#     Gary Saavedra (Sandia National Labs)
#     Peter Schultz (Sandia National Labs)
#     Laura Swiler (Sandia National Labs)
#
# <!-----------------END-HEADER------------------------------------->

from fitsnap3.io.input import config
from os import path, listdir, stat
import numpy as np
from random import shuffle
from fitsnap3.parallel_tools import pt
from fitsnap3.io.output import output
# from natsort import natsorted


def _float_to_int(a_float):
    if a_float == 0:
        return int(a_float)
    if a_float/int(a_float) != 1:
        raise ValueError("Training and Testing Size must be interpretable as integers")
    return int(a_float)


class Scraper:

    def __init__(self, name):
        self.name = name
        self.group_types = {}
        self.group_table = []
        self.files = {}
        self.configs = {}
        self.tests = None
        self.convert = {"Energy": 1.0, "Force": 1.0, "Stress": 1.0, "Distance": 1.0}
        self.data = {}

        self._init_units()

    def scrape_groups(self):
        group_dict = {k: config.sections["GROUPS"].group_types[i]
                      for i, k in enumerate(config.sections["GROUPS"].group_sections)}
        self.group_table = config.sections["GROUPS"].group_table
        size_type = None
        testing_size_type = None

        for key in self.group_table:
            bc_bool = False
            training_size = None
            if 'size' in self.group_table[key]:
                training_size = self.group_table[key]['size']
                bc_bool = True
                size_type = group_dict['size']
            if 'training_size' in self.group_table[key]:
                if training_size is not None:
                    raise ValueError("Do not set both size and training size")
                training_size = self.group_table[key]['training_size']
                size_type = group_dict['training_size']
            if 'testing_size' in self.group_table[key]:
                testing_size = self.group_table[key]['testing_size']
                testing_size_type = group_dict['testing_size']
            else:
                testing_size = 0
            if training_size is None:
                raise ValueError("Please set training size for {}".format(key))

            folder = path.join(config.sections["PATH"].datapath, key)
            folder_files = listdir(folder)
            for file_name in folder_files:
                if folder not in self.files:
                    self.files[folder] = []
                self.files[folder].append([folder + '/' + file_name, int(stat(folder + '/' + file_name).st_size)])
            shuffle(self.files[folder], pt.get_seed)
            nfiles = len(folder_files)
            if training_size < 1 or (training_size == 1 and size_type == float):
                if training_size == 1:
                    training_size = abs(training_size) * len(folder_files)
                else:
                    training_size = max(1, int(abs(training_size) * len(folder_files) - 0.5))
                if bc_bool and testing_size == 0:
                    testing_size = nfiles - training_size
            if testing_size != 0 and (testing_size < 1 or (testing_size == 1 and testing_size_type == float)):
                testing_size = max(1, int(abs(testing_size) * len(folder_files) - 0.5))
            training_size = _float_to_int(training_size)
            testing_size = _float_to_int(testing_size)
            if nfiles-testing_size-training_size < 0:
                raise ValueError("training size: {} + testing size: {} is greater than files in folder: {}".format(
                    training_size, testing_size, nfiles))
            output.screen(key, ": Detected ", nfiles, " fitting on ", training_size, " testing on ", testing_size)
            if self.tests is None:
                self.tests = {}
            self.tests[folder] = []
            for i in range(nfiles - training_size - testing_size):
                self.files[folder].pop()
            for i in range(testing_size):
                self.tests[folder].append(self.files[folder].pop())
            # self.files[folder] = natsorted(self.files[folder])

    # TODO : Fix divvy up to distribute groups evenly and based on memory
    def divvy_up_configs(self):
        groups = []
        group_list = []
        temp_list = []
        test_list = []
        for i, folder in enumerate(self.configs):
            for configuration in self.configs[folder]:
                if isinstance(configuration, list):
                    temp_list.append(configuration[0])
                else:
                    temp_list.append(configuration)
                groups.append(folder)

        self.configs = temp_list
        self.configs = pt.split_by_node(self.configs)

        if self.tests is not None:
            for i, folder in enumerate(self.tests):
                for file in self.tests[folder]:
                    test_list.append(file[0])
                    group_list.append(folder)
            test_list = pt.split_by_node(test_list)
            self.configs += test_list

        groups = pt.split_by_node(groups)
        group_list = pt.split_by_node(group_list)
        group_test = sorted(set(group_list))
        group_set = sorted(set(groups))
        group_counts = np.zeros((len(group_set) + len(group_test),), dtype='i')
        for i, group in enumerate(group_set+group_test):
            group_counts[i] = groups.count(group)

        pt.create_shared_array('files_per_group', len(self.configs), dtype='i')
        pt.shared_arrays['files_per_group'].array = group_counts

        pt.shared_arrays['files_per_group'].testing = 0
        if self.tests is not None:
            pt.shared_arrays['files_per_group'].testing = len(test_list)

        number_of_files_per_node = len(self.configs)
        pt.create_shared_array('number_of_atoms', number_of_files_per_node, dtype='i')
        pt.slice_array('number_of_atoms')
        self.configs = pt.split_within_node(self.configs)

    def scrape_configs(self):
        raise NotImplementedError("Call to virtual Scraper.scrape_configs method")

    def _init_units(self):
        if config.sections["REFERENCE"].units == "real":
            self.kb = 0.00198198665029335
        if config.sections["REFERENCE"].units == "metal":
            self.kb = 0.00008617333262145

    def _rotate_coords(self):
        # Transpose here because Lammps stores lattice vectors as columns,
        # QM stores lattice vectors as rows; After transposing lattice vectors are columns
        in_cell = np.asarray(self.data["QMLattice"]).T
        assert np.linalg.det(in_cell) > 0, "Input cell is not right-handed!"

        # Q matrix of QR decomposition is an orthogonal (rotation-like)
        # matrix whose inverse/transpose makes the input cell upper-diagonal:
        # input cell C = Q C';
        # runlammps-normalized cell C' = Q^T C.
        qmat, rmat = np.linalg.qr(in_cell)

        # Normalize signs of Q matrix to ensure positive diagonals of transformed cell;
        # QR decomposition algorithms don't always return a proper rotation
        ss = np.diagflat(np.sign(np.diag(rmat)))
        rot = ss @ qmat.T

        assert np.allclose(rot @ rot.T, np.eye(3)), "Rotation matrix not orthogonal"
        assert np.allclose(rot.T @ rot, np.eye(3)), "Rotation matrix not orthogonal"
        assert np.linalg.det(rot) > 0, "Rotation matrix is an improper rotation (det<0)"

        # ????
        # Cell transforms on first axis due to runlammps sotring lattice vectors as columns
        out_cell = rot @ in_cell

        # This assert is technically overkill, but checks that the new cell is right-handed
        assert np.linalg.det(out_cell) > 0, "New cell is not right-handed!"

        lower_triangle = out_cell[np.tril_indices(3, k=-1)]
        assert np.allclose(lower_triangle, 0, atol=1e-13), \
            f"Lower triangle of normalized cell has nonzero-elements: {lower_triangle}"

        # Positions and forces transform on the second axis
        # Stress transforms on both the first and second axis.
        self.data["Lattice"] = out_cell
        self.data["Positions"] = self.data["Positions"] * self.convert["Distance"] @ rot.T
        self.data["Forces"] = self.data["Forces"] * self.convert["Force"] @ rot.T
        self.data["Stress"] = rot @ (self.data["Stress"] * self.convert["Stress"]) @ rot.T
        self.data["Rotation"] = rot

    def _translate_coords(self):
        cell = self.data["Lattice"]
        position_in = self.data["Positions"]

        # Extra transposes because runlammps uses cells with latttice vectors as columns
        invcell = np.linalg.inv(cell.T).T
        # Fractional coordinates
        frac_coords = position_in @ invcell.T

        # Fix some rounding difficulties in divmod when within machine epsilon of zero
        frac_coords[np.isclose(frac_coords, 0, atol=1e-15)] = 0.

        trans_nums, cell_frac_coords = np.divmod(frac_coords, 1)

        assert (cell_frac_coords < 1).all(), "Fractional coordinates outside cell"
        assert (cell_frac_coords >= 0).all(), "Fractional coordinates outside cell"

        # If no translations are needed, return unmodified positions
        if (trans_nums == 0).all():
            self.data["Positions"] = position_in
            self.data["Translation"] = np.zeros_like(position_in, dtype=float)

        new_pos = cell_frac_coords @ cell.T
        trans_vec = trans_nums @ cell.T
        assert np.allclose(new_pos + trans_vec, position_in), "Translation failed to invert"
        self.data["Postions"] = new_pos
        self.data["Translation"] = trans_vec

    def _stress_conv(self, styles):

        if (config.sections["REFERENCE"].units == "metal" and
                list(styles["Stress"])[0] == "kbar" or list(styles["Stress"])[0] == "kB"):
            self.convert["Stress"] = 1000.0

    # def check_coords(self, cell, pos1, pos2):
    #     """Compares position 1 and position 2 with respect to periodic boundaries defined by cell"""
    #     invcell = np.linalg.inv(np.asarray(cell).T).T
    #
    #     # Fractional coordinates
    #     frac_1 = pos1 @ invcell.T
    #     frac_2 = pos2 @ invcell.T
    #     diff_frac = frac_2 - frac_1
    #
    #     # Assert that diff_frac is very close to an integer
    #     assert np.allclose(
    #         diff_frac,
    #         np.round(diff_frac),
    #         atol=1e-12, rtol=1e-12), "Coordinates are not close after shift." + \
    #                                  "Fractional coordinate Error:{}\nArray:\n{}".format(
    #                                      np.abs(diff_frac).max(), diff_frac)
    #     return True
    #
    # def check_volume(self, lattice, volume):
    #     assert np.allclose(np.linalg.det(lattice), volume, rtol=1e-10, atol=1e-10), \
    #         "Cell volume not equal to supplied volume!"



# <!----------------BEGIN-HEADER------------------------------------>
# ## FitSNAP3
# A Python Package For Training SNAP Interatomic Potentials for use in the LAMMPS molecular dynamics package
#
# _Copyright (2016) Sandia Corporation.
# Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
# the U.S. Government retains certain rights in this software.
# This software is distributed under the GNU General Public License_
# ##
#
# #### Original author:
#     Aidan P. Thompson, athomps (at) sandia (dot) gov (Sandia National Labs)
#     http://www.cs.sandia.gov/~athomps
#
# #### Key contributors (alphabetical):
#     Mary Alice Cusentino (Sandia National Labs)
#     Nicholas Lubbers (Los Alamos National Lab)
#     Charles Sievers (Sandia National Labs)
#     Adam Stephens (Sandia National Labs)
#     Mitchell Wood (Sandia National Labs)
#
# #### Additional authors (alphabetical):
#     Elizabeth Decolvenaere (D. E. Shaw Research)
#     Stan Moore (Sandia National Labs)
#     Steve Plimpton (Sandia National Labs)
#     Gary Saavedra (Sandia National Labs)
#     Peter Schultz (Sandia National Labs)
#     Laura Swiler (Sandia National Labs)
#
# <!-----------------END-HEADER------------------------------------->

from time import time, sleep
import numpy as np
from lammps import lammps
from random import random
from psutil import virtual_memory
from itertools import chain
import ctypes
import signal
try:
    # stubs = 0 MPI is active
    stubs = 0
    from mpi4py import MPI
except ModuleNotFoundError:
    stubs = 1


def printf(*args, **kw):
    print(*args, flush=True)


class GracefulKiller:

    def __init__(self, comm, failsafe=False):
        self._comm = comm
        self._rank = 0
        self.failsafe = failsafe
        if self._comm is not None:
            self._rank = self._comm.Get_rank()
            signal.signal(signal.SIGINT, self.exit_gracefully)
            signal.signal(signal.SIGTERM, self.exit_gracefully)

    def exit_gracefully(self, signum, frame):
        if self._rank == 0:
            printf("attempting to exit gracefully")
            printf("exiting from exit code", signum, "at", frame)
        if self.failsafe:
            self._comm.Abort()


def _rank_zero(method):
    def check_if_rank_zero(*args, **kw):
        if args[0].get_rank() == 0:
            return method(*args, **kw)
        else:
            return dummy_function()
    return check_if_rank_zero


def _sub_rank_zero(method):
    def check_if_rank_zero(*args, **kw):
        if args[0].get_rank() == 0:
            return method(*args, **kw)
        else:
            return dummy_function()
    return check_if_rank_zero


def identity_decorator(self, obj):
    return obj


def _rank_zero_decorator(decorator):
    def check_if_rank_zero(*args, **kw):
        if args[0].get_rank() == 0:
            return decorator(*args, **kw)
        else:
            return identity_decorator(*args, **kw)
    return check_if_rank_zero


def dummy_function(*args, **kw):
    return None


def stub_check(method):
    def stub_function(*args, **kw):
        if stubs == 0:
            return method(*args, **kw)
        else:
            return dummy_function(*args, **kw)
    return stub_function


def print_lammps(method):
    def new_method(*args, **kw):
        printf(*args)
        return method(*args, **kw)
    return new_method


class ParallelTools:

    def __init__(self):
        if stubs == 0:
            self._comm = MPI.COMM_WORLD
            self._rank = self._comm.Get_rank()
            self._size = self._comm.Get_size()
        if stubs == 1:
            self._rank = 0
            self._size = 1
            self._comm = None
            self._sub_rank = 0
            self._sub_size = 1
            self._sub_comm = None
            self._sub_head_proc = 0
            self._node_index = 0
            self._number_of_nodes = 1

        killer = GracefulKiller(self._comm)

        self._comm_split()
        self._lmp = None
        self._seed = 0.0
        self._set_seed()
        self.shared_arrays = {}
        self._calculator_options = {}
        self.logger = None

    @stub_check
    def _comm_split(self):
        self._sub_comm = self._comm.Split_type(MPI.COMM_TYPE_SHARED)
        self._sub_rank = self._sub_comm.Get_rank()
        self._sub_size = self._sub_comm.Get_size()
        self._sub_head_proc = 0
        if self._sub_rank == 0:
            self._sub_head_proc = self._rank
        self._sub_head_proc = self._sub_comm.bcast(self._sub_head_proc)
        self._sub_head_procs = list(dict.fromkeys(self._comm.allgather(self._sub_head_proc)))
        self._head_group = self._comm.group.Incl(self._sub_head_procs)
        self._head_group_comm = self._comm.Create_group(self._head_group)
        self._node_index = self._sub_head_procs.index(self._sub_head_proc)
        self._number_of_nodes = len(self._sub_head_procs)
        self._micro_comm = self._comm.Split(self._rank)

    def _set_seed(self):
        if self._rank == 0.0:
            self._seed = random()
        if stubs == 0:
            self._seed = self._comm.bcast(self._seed)

    def get_seed(self):
        return self._seed

    def get_rank(self):
        return self._rank

    @_rank_zero
    def single_print(self, *args, **kw):
        printf(*args)

    @_sub_rank_zero
    def sub_print(self, *args, **kw):
        printf(*args)

    def all_print(self, *args, **kw):
        printf("Rank", self._rank, ":", *args)

    @_rank_zero_decorator
    def single_timeit(self, method):
        def timed(*args, **kw):
            ts = time()
            result = method(*args, **kw)
            te = time()
            if 'log_time' in kw:
                name = kw.get('log_name', method.__name__.upper())
                kw['log_time'][name] = int((te - ts) * 1000)
            else:
                printf("'{0}' took {1:.2f} ms on rank {2}".format(
                    method.__name__, (te - ts) * 1000, self._rank))
            return result
        return timed

    def per_rank_timeit(self, method):
        def timed(*args, **kw):
            ts = time()
            result = method(*args, **kw)
            te = time()
            if 'log_time' in kw:
                name = kw.get('log_name', method.__name__.upper())
                kw['log_time'][name] = int((te - ts) * 1000)
            else:
                printf("'{0}' took {1:.2f} ms on rank {2}".format(
                    method.__name__, (te - ts) * 1000, self._rank))
            return result
        return timed

    def rank_zero(self, method):
        if self._rank == 0:
            def check_if_rank_zero(*args, **kw):
                return method(*args, **kw)
            return check_if_rank_zero
        else:
            return dummy_function

    def sub_rank_zero(self, method):
        if self._sub_rank == 0:
            def check_if_rank_zero(*args, **kw):
                return method(*args, **kw)
            return check_if_rank_zero
        else:
            return dummy_function

    def create_shared_array(self, name, size1, size2=1, dtype='d'):

        if isinstance(name, str):
            if stubs == 0:
                self.shared_arrays[name] = SharedArray(size1, size2=size2,
                                                       dtype=dtype,
                                                       comm=self._sub_comm,
                                                       rank=self._sub_rank)
            else:
                self.shared_arrays[name] = StubsArray(size1, size2, dtype=dtype)
        else:
            raise TypeError("name must be a string")

    @stub_check
    def allgather(self, array):
        return self._head_group_comm.allgather(array)

    def add_calculator_option(self, option, boolean):
        self._calculator_options[option] = boolean

    @stub_check
    def all_barrier(self):
        self._comm.Barrier()

    @stub_check
    def sub_barrier(self):
        self._sub_comm.Barrier()

    def split_by_node(self, obj):
        if isinstance(obj, list):
            return obj[self._node_index::self._number_of_nodes]
        elif isinstance(obj, dict):
            for key in obj:
                obj[key] = obj[key][self._node_index::self._number_of_nodes]
            return obj
        elif isinstance(obj, np.ndarray):
            return obj[self._node_index::self._number_of_nodes]
        else:
            raise TypeError("Parallel tools cannot split {} by node.".format(obj))

    def split_within_node(self, obj):
        if isinstance(obj, list):
            return obj[self._sub_rank::self._sub_size]
        elif isinstance(obj, dict):
            for key in obj:
                obj[key] = obj[key][self._node_index::self._number_of_nodes]
            return obj
        else:
            raise TypeError("Parallel tools cannot split {} within node.".format(obj))

    def initialize_lammps(self, lammpslog=0, printlammps=0):
        cmds = ["-screen", "none"]
        if not lammpslog:
            cmds.append("-log")
            cmds.append("none")
        if stubs == 0:
            self._lmp = lammps(comm=self._micro_comm, cmdargs=cmds)
        else:
            self._lmp = lammps(cmdargs=cmds)

        if printlammps == 1:
            self._lmp.command = print_lammps(self._lmp.command)
        return self._lmp

    def close_lammps(self):
        if self._lmp is not None:
            # Kill lammps jobs
            self._lmp.close()
            self._lmp = None
        return self._lmp

    def slice_array(self, name, num_types=None):
        if name in self.shared_arrays:
            if name != 'a':
                s = slice(self._sub_rank, None, self._sub_size)
                self.shared_arrays[name].sliced_array = self.shared_arrays[name].array[s][:]
            else:
                self.slice_a()
        else:
            raise IndexError("{} not found in shared objects".format(name))

    def slice_a(self, name='a'):
        nof = len(pt.shared_arrays["number_of_atoms"].array)
        s = slice(self._sub_rank, nof, self._sub_size)
        indices = [0]
        self.shared_arrays[name].group_index = []
        self.shared_arrays[name].group_energy_index = []
        self.shared_arrays[name].group_force_index = []
        self.shared_arrays[name].group_stress_index = []
        count = [0]
        for i, value in enumerate(self.shared_arrays['files_per_group'].array):
            count.append(count[i]+value)
        j = 0
        e_temp = []
        f_temp = []
        s_temp = []
        atoms = []
        for i in range(nof):
            if i in count:
                self.shared_arrays[name].group_index.append(j)
                if self._calculator_options["energy"] and i > 0:
                    self.shared_arrays[name].group_energy_index.append(e_temp)
                    e_temp = []
                if self._calculator_options["force"] and i > 0:
                    self.shared_arrays[name].group_force_index.append(f_temp)
                    f_temp = []
                if self._calculator_options["stress"] and i > 0:
                    self.shared_arrays[name].group_stress_index.append(s_temp)
                    s_temp = []
            if self._calculator_options["energy"]:
                e_temp.append(j)
                j += 1
            if self._calculator_options["force"]:
                f_temp.append(j)
                j += 3 * pt.shared_arrays["number_of_atoms"].array[i]
            if self._calculator_options["stress"]:
                s_temp.append(j)
                j += 6
            indices.append(j)
            atoms.append(pt.shared_arrays["number_of_atoms"].array[i])
        if self._calculator_options["energy"]:
            self.shared_arrays[name].group_energy_index.append(e_temp)
            self.shared_arrays[name].energy_index = \
                list(chain.from_iterable(self.shared_arrays[name].group_energy_index))
            self.shared_arrays[name].group_energy_length = \
                sum(len(row) for row in self.shared_arrays[name].group_energy_index)
        if self._calculator_options["force"]:
            self.shared_arrays[name].group_force_index.append(f_temp)
            self.shared_arrays[name].force_index = \
                list(chain.from_iterable(self.shared_arrays[name].group_force_index))
            self.shared_arrays[name].group_force_length = \
                sum(len(row) for row in self.shared_arrays[name].group_force_index)
        if self._calculator_options["stress"]:
            self.shared_arrays[name].group_stress_index.append(s_temp)
            self.shared_arrays[name].stress_index = \
                list(chain.from_iterable(self.shared_arrays[name].group_stress_index))
            self.shared_arrays[name].group_stress_length = \
                sum(len(row) for row in self.shared_arrays[name].group_stress_index)
        self.shared_arrays[name].group_index.append(j)
        self.shared_arrays[name].indices = indices[s]
        self.shared_arrays[name].num_atoms = atoms

    @stub_check
    def combine_coeffs(self, coeff):
        new_coeff = None
        if self._sub_rank == 0:
            new_coeff = self._head_group_comm.allreduce(coeff) / self._number_of_nodes
        return self._sub_comm.bcast(new_coeff)

    @staticmethod
    def get_ram():
        mem = virtual_memory()
        return mem.total

    def set_logger(self, logger):
        self.logger = logger

    def abort(self):
        self._comm.Abort()

    def exception(self, err):
        if pt.logger is None and self._rank == 0:
            raise err

        pt.close_lammps()
        if self._rank == 0:
            self.logger.exception(err)
        sleep(5)
        self.abort()


class SharedArray:

    def __init__(self, size1, size2=1, dtype='d', comm=None, rank=0):

        # total array for all procs
        self.array = None
        # sub array for this proc
        self.sliced_array = None

        self.energies_index = None
        self.forces_index = None
        self.strain_index = None

        if dtype == 'd':
            item_size = MPI.DOUBLE.Get_size()
        elif dtype == 'i':
            item_size = MPI.INT.Get_size()
        else:
            raise TypeError("dtype {} has not been implemented yet".format(dtype))
        if rank == 0:
            self._nbytes = size1 * size2 * item_size
        else:
            self._nbytes = 0

        win = MPI.Win.Allocate_shared(self._nbytes, item_size, comm=comm)

        buff, item_size = win.Shared_query(0)

        if dtype == 'd':
            assert item_size == MPI.DOUBLE.Get_size()
        elif dtype == 'i':
            assert item_size == MPI.INT32_T.Get_size()
        if size2 == 1:
            self.array = np.ndarray(buffer=buff, dtype=dtype, shape=(size1, ))
        else:
            self.array = np.ndarray(buffer=buff, dtype=dtype, shape=(size1, size2))

    def get_memory(self):
        return self._nbytes


class StubsArray:

    def __init__(self, size1, size2=1, dtype='d'):
        # total array for all procs
        self.array = None
        # sub array for this proc
        self.sliced_array = None

        self.energies_index = None
        self.forces_index = None
        self.strain_index = None

        if size2 == 1:
            self.array = np.ndarray(shape=(size1, ), dtype=dtype)
        else:
            self.array = np.ndarray(shape=(size1, size2), dtype=dtype)

    def get_memory(self):
        return self.array.nbytes


class Output:

    def __init__(self):
        self.none = None

    @staticmethod
    def screen(*args, **kw):
        pt.single_print(*args, **kw)

    @staticmethod
    def exception(err):
        # There is almost never a reason to use this
        pt.exception(err)


if __name__ == "fitsnap3.parallel_tools":
    pt = ParallelTools()
    if stubs == 0:
        double_size = MPI.DOUBLE.Get_size()
    else:
        double_size = ctypes.sizeof(ctypes.c_double)
    output = Output()



from fitsnap3.calculators.calculator import Calculator
"""Methods you may or must override in new calculators"""


class Template(Calculator):

    def __init__(self, name):
        super().__init__(name)

    # Calculator must override process_configs method
    def process_configs(self, data, i):
        """"""
        pass



"""Groups Class"""
from fitsnap3.io.sections.sections import Section, output
from pandas import read_csv


def _str_2_fun(some_list):
    for i, item in enumerate(some_list):
        if item == 'str':
            some_list[i] = str
        if item == 'bool':
            some_list[i] = bool
        if item == 'int':
            some_list[i] = int
        if item == 'float':
            some_list[i] = float


class Groups(Section):

    def __init__(self, name, config, args):
        super().__init__(name, config, args)
        self.group_sections = self.get_value("GROUPS", "group_sections", "name size eweight fweight vweight").split()
        self.group_types = self.get_value("GROUPS", "group_types", "str float float float float").split()
        _str_2_fun(self.group_types)
        self.group_table = None
        if self.get_value("PATH", "groupFile", "None") is not "None":
            self.read_group_file()
        else:
            self.read_group_config()
        # run init methods to populate class
        # delete config and args to rely only on child's members
        self.delete()

    def read_group_config(self):
        self.group_table = {k: v.split() for (k, v) in self.get_section("GROUPS")}
        # Deletes any key:value from self.groups that shares a key name with the vars of this instance
        for k in vars(self):
            if k in self.group_table:
                self.group_table.pop(k)
        for k, v in self.group_table.items():
            self.group_table[k] = {self.group_sections[i+1]: self.group_types[i+1](item) for i, item in enumerate(v)}

    def read_group_file(self):
        group_types = {self.group_sections[i]: item for i, item in enumerate(self.group_types)}
        group_table = read_csv(self.get_value("PATH", "groupFile", "None"),
                               delim_whitespace=True,
                               comment='#',
                               skip_blank_lines=True,
                               names=self.group_sections,
                               index_col=False)

        # Remove blank lines ; skip_blank_lines doesn't seem to work.
        group_table = group_table.dropna()
        group_table.index = range(len(group_table.index))

        # Convert data types
        group_table = group_table.astype(dtype=group_types)
        self.group_table = {}
        for data_frame in group_table.itertuples():
            self.group_table[data_frame[1]] = {self.group_sections[i+1]: item for i, item in enumerate(data_frame[2:])}



from fitsnap3.solvers.solver import Solver
from fitsnap3.parallel_tools import pt
from scipy.linalg import lstsq
import numpy as np


class SVD(Solver):

    def __init__(self, name):
        super().__init__(name)

    def perform_fit(self):
        if pt.shared_arrays['files_per_group'].testing != 0:
            testing = -1*pt.shared_arrays['files_per_group'].testing
        else:
            testing = len(pt.shared_arrays['w'].array)
        w = pt.shared_arrays['w'].array[:testing]
        aw, bw = w[:, np.newaxis] * pt.shared_arrays['a'].array[:testing], w * pt.shared_arrays['b'].array[:testing]
        self.fit, residues, rank, s = lstsq(aw, bw, 1.0e-13)



import configparser
import argparse
from pickle import HIGHEST_PROTOCOL
from fitsnap3.io.sections.section_factory import new_section
from fitsnap3.parallel_tools import pt, output
from os import path, listdir


class Config:

    def __init__(self):
        self.default_protocol = HIGHEST_PROTOCOL
        self.args = None
        self.parse_cmdline()
        self.sections = {}
        self.parse_config()
        self._reset_chemflag()

    def parse_cmdline(self):
        parser = argparse.ArgumentParser(prog="FitSNAP3")

        parser.add_argument("infile", action="store",
                            help="Input file with bispectrum etc. options")
        # Not Implemented
        parser.add_argument("--verbose", "-v", action="store_true", dest="verbose",
                            default=False, help="Show more detailed information about processing")
        parser.add_argument("--lammpslog", "-l", action="store_true", dest="lammpslog",
                            help="Write logs from LAMMPS. Logs will appear in current working directory.")
        parser.add_argument("--printlammps", "-pl", action="store_true", dest="printlammps",
                            help="Print all lammps commands")
        parser.add_argument("--relative", "-r", action="store_true", dest="relative",
                            help="""Put output files in the directory of INFILE. If this flag
                             is not not present, the files are stored in the
                            current working directory.""")
        # Not Implemented
        parser.add_argument("--nofit", "-nf", action="store_false", dest="perform_fit",
                            help="Don't perform fit, just compute bispectrum data.")
        # Not Implemented
        parser.add_argument("--overwrite", action="store_true", dest="overwrite",
                            help="Allow overwriting existing files")
        # Not Implemented
        parser.add_argument("--lammps_noexceptions", action="store_true",
                            help="Allow LAMMPS compiled without C++ exceptions handling enabled")
        # Not Implemented
        parser.add_argument("--keyword", "-k", nargs=3, metavar=("GROUP", "NAME", "VALUE"),
                            action="append", dest="keyword_replacements",
                            help="""Replace or add input keyword group GROUP, key NAME,
                            with value VALUE. Type carefully; a misspelled key name or value
                            may be silently ignored.""")
        parser.add_argument("--screen", "-sc", action="store_false", dest="screen",
                            help="Print to screen")
        parser.add_argument("--nscreen", "-ns", action="store_true", dest="nscreen",
                            help="Print each nodes screen")
        parser.add_argument("--pscreen", "-ps", action="store_true", dest="pscreen",
                            help="Print each processors screen")
        parser.add_argument("--log", action="store", dest="log",
                            default=None, help="Write fitsnap log to this file")

        self.args = parser.parse_args()

    def parse_config(self):

        tmp_config = configparser.ConfigParser(inline_comment_prefixes='#')
        tmp_config.optionxform = str
        tmp_config.read(self.args.infile)

        vprint = output.screen if self.args.verbose else lambda *arguments, **kwargs: None
        if self.args.keyword_replacements:
            for kwg, kwn, kwv in self.args.keyword_replacements:
                if kwg not in tmp_config:
                    raise ValueError(f"{kwg} is not a valid keyword group")
                vprint(f"Substituting {kwg}:{kwn}={kwv}")
                tmp_config[kwg][kwn] = kwv

        self.set_sections(tmp_config)

    def set_sections(self, tmp_config):
        location = '/' + '/'.join(path.abspath(__file__).split("/")[:-1]) + '/sections'
        temp = {file: None for file in listdir(location)}
        for file in temp:
            if file.split('.')[-1] != 'py' or file == 'sections.py' or file == 'section_factory.py':
                continue
            else:
                section = '.'.join(file.split('.')[:-1]).upper()
                if section == "TEMPLATE":
                    section = "DEFAULT"
                self.sections[section] = new_section(section, tmp_config, self.args)
        del temp

    def _reset_chemflag(self):
        if self.sections["CALCULATOR"].chemflag != 0:
            chemflag = "{}".format(self.sections['BISPECTRUM'].numtypes)
            for element in self.sections["BISPECTRUM"].type_mapping:
                element_type = self.sections["BISPECTRUM"].type_mapping[element]
                chemflag += " {}".format(element_type - 1)
            self.sections["CALCULATOR"].chemflag = "{}".format(chemflag)


if __name__ == "fitsnap3.io.input":
    config = Config()



from fitsnap3.io.sections.sections import Section
from os import getcwd, path


class Path(Section):

    def __init__(self, name, config, args):
        super().__init__(name, config, args)
        paths = getcwd().split('/')+args.infile.split('/')[:-1]
        working_directory = ''
        for directory in paths[:-1]:
            working_directory += directory + '/'
        working_directory += paths[-1]
        self.datapath = path.join(working_directory, self.get_value("PATH", "dataPath", "JSON"))
        self.group_file = path.join(working_directory, self.get_value("PATH", "groupFile", "grouplist.in"))
        self.smartweights = self.get_value("PATH", "smartweights", "0", "bool")
        self.delete()



import setuptools

with open("fitsnap3/README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
     name='fitsnap3',  
     version='0.9',
     author="Charles Sievers",
     author_email="charliesievers@cox.net",
     description="Interatomic Potential Machine learning Interface",
     long_description=long_description,
     long_description_content_type="text/markdown",
     url="https://github.com/charlessievers/FitSNAP/tree/ObjOrPar/fitsnap3",
     packages=setuptools.find_packages(),
     install_requires=["psutil>=5.6.3", "scipy>=1.3.1", "pandas>=0.25.1", "numpy>=1.17.2"],
     package_data={
         "fitsnap3": ["*/*.py", "*/*/*.py"]
     },
     classifiers=[
         "Programming Language :: Python :: 3",
         "License :: OSI Approved :: MIT License",
         "Operating System :: OS Independent",
     ],
 )



from fitsnap3.scrapers.scrape import Scraper
from fitsnap3.io.input import config
from json import loads
from fitsnap3.parallel_tools import pt
from fitsnap3.io.output import output
import numpy as np
from _collections import defaultdict


class Json(Scraper):

    def __init__(self, name):
        super().__init__(name)
        self.style_vars = ['AtomType', 'Stress', 'Lattice', 'Energy', "Positions", "Forces"]
        self.array_vars = ['AtomTypes', 'Stress', 'Lattice', "Positions", "Forces"]
        if config.sections["REFERENCE"].atom_style == "spin":
            self.style_vars.append("Spins")
            self.array_vars.append("Spins")
        if config.sections["REFERENCE"].atom_style == "charge":
            self.style_vars.append("Charges")
            self.array_vars.append("Charges")
        self.styles = defaultdict(lambda: set())
        self.all_data = []
        self.style_info = {}

    def scrape_groups(self):
        super().scrape_groups()
        self.configs = self.files

    def scrape_configs(self):
        self.files = self.configs
        for i, file_name in enumerate(self.files):
            with open(file_name) as file:
                file.readline()
                try:
                    self.data = loads(file.read(), parse_constant=True)
                except Exception as e:
                    output.screen("Trouble Parsing Training Data: ", file_name)
                    output.exception(e)

                assert len(self.data) == 1, "More than one object (dataset) is in this file"

                self.data = self.data['Dataset']

                assert len(self.data['Data']) == 1, "More than one configuration in this dataset"

                self.data['Group'] = file_name.split("/")[-2]
                self.data['File'] = file_name.split("/")[-1]

                for sty in self.style_vars:
                    self.styles[sty].add(self.data.pop(sty + "Style", ))

                assert all(k not in self.data for k in self.data["Data"][0].keys()), \
                    "Duplicate keys in dataset and data"

                self.data.update(self.data.pop('Data')[0])  # Move data up one level

                for key in self.array_vars:
                    self.data[key] = np.asarray(self.data[key])

                natoms = np.shape(self.data["Positions"])[0]
                pt.shared_arrays["number_of_atoms"].sliced_array[i] = natoms
                self.data["QMLattice"] = self.data["Lattice"]
                del self.data["Lattice"]  # We will populate this with the lammps-normalized lattice.
                if "Label" in self.data:
                    del self.data["Label"]  # This comment line is not that useful to keep around.

                # possibly due to JSON, some configurations have integer energy values.
                if not isinstance(self.data["Energy"], float):
                    # pt.print(f"Warning: Configuration {all_index}
                    # ({group_name}/{fname_end}) gives energy as an integer")
                    self.data["Energy"] = float(self.data["Energy"])

                self._stress_conv(self.styles)
                self.data["Energy"] *= self.convert["Energy"]

                if hasattr(config.sections["ESHIFT"], 'eshift'):
                    for atom in self.data["AtomTypes"]:
                        self.data["Energy"] += config.sections["ESHIFT"].eshift[atom]

                self._rotate_coords()
                self._translate_coords()

                if self.group_table[self.data['Group']]['eweight'] >= 0.0:
                    for key in self.group_table[self.data['Group']]:
                        # Do not put the word weight in a group table unless you want to use it as a weight
                        if 'weight' in key:
                            self.data[key] = self.group_table[self.data['Group']][key]
                else:
                    self.data['eweight'] = np.exp(
                        (self.group_table[self.data['Group']]['eweight'] - self.data["Energy"] /
                         float(natoms)) / (self.kb * float(config.sections["BISPECTRUM"].boltz)))
                    self.data['fweight'] = \
                        self.data['eweight'] * self.group_table[self.data['Group']]['fweight']
                    self.data['vweight'] = \
                        self.data['eweight'] * self.group_table[self.data['Group']]['vweight']

                self.all_data.append(self.data)

        for style_name, style_set in self.styles.items():
            assert len(style_set) == 1, "Multiple styles ({}) for {}".format(len(style_set), style_name)

        self.style_info = {k: v.pop() for k, v in self.styles.items()}

        return self.all_data



from fitsnap3.io.sections.sections import Section


class Reference(Section):

    def __init__(self, name, config, args):
        super().__init__(name, config, args)
        self.units = self.get_value("REFERENCE", "units", "metal").lower()
        self.atom_style = self.get_value("REFERENCE", "atom_style", "atomic").lower()
        self.lmp_pairdecl = []
        self.lmp_pairdecl.append("pair_style " + self.get_value("REFERENCE", "pair_style", "zero 10.0"))
        for name, value in self._config.items("REFERENCE"):
            if not name.find("pair_coeff"):
                self.lmp_pairdecl.append("pair_coeff " + value)
        if "pair_coeff" in self.lmp_pairdecl:
            self.lmp_pairdecl.append("pair_coeff * * ")
        self.delete()



from fitsnap3.io.input import config
from fitsnap3.parallel_tools import pt
import numpy as np
from pandas import DataFrame


class Solver:

    def __init__(self, name):
        self.name = name
        self.configs = None
        self.fit = None
        self.all_fits = None
        self.template_error = False
        self.errors = []
        self.weighted = 'Unweighted'
        self.residuals = None
        self.a = None
        self.b = None
        self.w = None

    def perform_fit(self):
        pass

    def fit_gather(self):
        self.all_fits = pt.allgather(self.fit)

    def _offset(self):
        num_types = config.sections["BISPECTRUM"].numtypes
        if num_types > 1:
            self.fit = self.fit.reshape(num_types, config.sections["BISPECTRUM"].ncoeff)
            offsets = np.zeros((num_types, 1))
            self.fit = np.concatenate([offsets, self.fit], axis=1)
            self.fit = self.fit.reshape((-1, 1))
        else:
            self.fit = np.insert(self.fit, 0, 0)

    def error_analysis(self, data):
        for option in ["Unweighted", "Weighted"]:
            self.weighted = option
            self._all_error()
            self._group_error(data)

        if self.template_error is True:
            self._template_error()

        self.errors = DataFrame.from_records(self.errors)
        self.errors = self.errors.set_index(["Group", "Weighting", "Subsystem", ]).sort_index()

        if config.sections["CALCULATOR"].bzeroflag:
            self._offset()

    def _all_error(self):
        if config.sections["CALCULATOR"].energy:
            self._energy()
        if config.sections["CALCULATOR"].force:
            self._force()
        if config.sections["CALCULATOR"].stress:
            self._stress()
        self._combined()

    def _energy(self):
        testing = -1 * pt.shared_arrays['files_per_group'].testing
        a, b, w = self._make_abw(pt.shared_arrays['a'].energy_index, 1)
        self._errors([[0, testing]], ['*ALL'], "Energy", a, b, w)
        if testing != 0:
            self._errors([[testing, 0]], ['*ALL'], "Energy_testing", a, b, w)

    def _force(self):
        testing = -1 * pt.shared_arrays['files_per_group'].testing
        a, b, w = self._make_abw(pt.shared_arrays['a'].force_index, pt.shared_arrays['a'].num_atoms)
        self._errors([[0, testing]], ['*ALL'], "Force", a, b, w)
        if testing != 0:
            self._errors([[testing, 0]], ['*ALL'], "Force_testing", a, b, w)

    def _stress(self):
        testing = -1 * pt.shared_arrays['files_per_group'].testing
        a, b, w = self._make_abw(pt.shared_arrays['a'].stress_index, 6)
        self._errors([[0, testing]], ['*ALL'], "Stress", a, b, w)
        if testing != 0:
            self._errors([[testing, 0]], ['*ALL'], "Stress_testing", a, b, w)

    @staticmethod
    def _make_abw(type_index, buffer):
        if isinstance(buffer, list):
            length = sum(buffer)
            length *= 3
        else:
            length = len(type_index) * buffer
        width = np.shape(pt.shared_arrays['a'].array)[1]
        a = np.zeros((length, width))
        b = np.zeros((length,))
        w = np.zeros((length,))
        i = 0
        for j, value in enumerate(type_index):
            if isinstance(buffer, list):
                spacing = buffer[j]*3
            else:
                spacing = buffer
            a[i:i+spacing] = pt.shared_arrays['a'].array[value:value+spacing]
            b[i:i+spacing] = pt.shared_arrays['b'].array[value:value+spacing]
            w[i:i+spacing] = pt.shared_arrays['w'].array[value:value+spacing]
            i += spacing
        return a, b, w

    def _combined(self):
        testing = -1 * pt.shared_arrays['files_per_group'].testing
        self._errors([[0, testing]], ["*ALL"], "Combined")
        if testing != 0:
            self._errors([[testing, 0]], ['*ALL'], "Combined_testing")

    def _group_error(self, data):
        groups = []
        testing = len(data)-pt.shared_arrays['files_per_group'].testing
        for i, file in enumerate(data):
            if i < testing:
                groups.append(file["Group"])
            else:
                groups.append(file["Group"]+'_Testing')
        groups = sorted(set(groups))
        if config.sections["CALCULATOR"].energy:
            self._group_energy(groups)
        if config.sections["CALCULATOR"].force:
            self._group_force(groups)
        if config.sections["CALCULATOR"].stress:
            self._group_stress(groups)
        self._group_combined(groups)

    def _group_energy(self, groups):
        group_index = pt.shared_arrays['a'].group_energy_index
        length = pt.shared_arrays['a'].group_energy_length
        index, a, b, w = self._make_group_abw(group_index, length, 1)
        self._errors(index, groups, "Energy", a, b, w)

    def _group_force(self, groups):
        group_index = pt.shared_arrays['a'].group_force_index
        length = pt.shared_arrays['a'].group_force_length
        index, a, b, w = self._make_group_abw(group_index, length, pt.shared_arrays['a'].num_atoms)
        self._errors(index, groups, "Force", a, b, w)

    def _group_stress(self, groups):
        group_index = pt.shared_arrays['a'].group_stress_index
        length = pt.shared_arrays['a'].group_stress_length*6
        index, a, b, w = self._make_group_abw(group_index, length, 6)
        self._errors(index, groups, "Stress", a, b, w)

    @staticmethod
    def _make_group_abw(group_index, length, buffer):
        index = []
        width = np.shape(pt.shared_arrays['a'].array)[1]
        if isinstance(buffer, list):
            length = 3 * sum(buffer)
        a = np.zeros((length, width))
        b = np.zeros((length,))
        w = np.zeros((length,))
        i = 0
        j = 0
        for group in group_index:
            temp = [i]
            for value in group:
                if isinstance(buffer, list):
                    spacing = buffer[j]
                    j += 1
                    spacing *= 3
                else:
                    spacing = buffer
                a[i:i+spacing] = pt.shared_arrays['a'].array[value:value+spacing]
                b[i:i+spacing] = pt.shared_arrays['b'].array[value:value+spacing]
                w[i:i+spacing] = pt.shared_arrays['w'].array[value:value+spacing]
                i += spacing
            temp.append(i)
            index.append(temp)
        return index, a, b, w

    def _group_combined(self, groups):
        index = []
        group_index = pt.shared_arrays['a'].group_index
        for i in range(len(group_index) - 1):
            index.append([group_index[i], group_index[i+1]])
        self._errors(index, groups, "Combined")

    def _errors(self, index, category, gtype, a=None, b=None, w=None):
        if a is None:
            a = pt.shared_arrays['a'].array
        if b is None:
            b = pt.shared_arrays['b'].array
        if w is None:
            w = pt.shared_arrays['w'].array
        for i, group in enumerate(category):
            if index[i][1] != 0:
                a_err = a[index[i][0]:index[i][1]]
                b_err = b[index[i][0]:index[i][1]]
                w_err = w[index[i][0]:index[i][1]]
            else:
                a_err = a[index[i][0]:]
                b_err = b[index[i][0]:]
                w_err = w[index[i][0]:]
            true, pred = b_err, a_err @ self.fit
            if self.weighted == 'Weighted':
                true, pred = w_err * true, w_err * pred
                nconfig = np.count_nonzero(w_err)
            else:
                nconfig = len(pred)
            res = true - pred
            mae = np.sum(np.abs(res) / nconfig)
            mean_dev = np.sum(np.abs(true - np.median(true)) / nconfig)
            ssr = np.square(res).sum()
            mse = ssr / nconfig
            rmse = np.sqrt(mse)
            rsq = 1 - ssr / np.sum(np.square(true - (true / nconfig).sum()))
            error_record = {
                "Group": group,
                "Weighting": self.weighted,
                "Subsystem": gtype,
                "ncount": nconfig,
                "mae": mae,
                "rmae": mae / mean_dev,
                "rmse": rmse,
                "rrmse": rmse / np.std(true),
                "ssr": ssr,
                "rsq": rsq
            }
            if self.residuals is not None:
                error_record["residual"] = res
            self.errors.append(error_record)

    def _template_error(self):
        pass



from fitsnap3.io.input import config
from fitsnap3.parallel_tools import pt
from contextlib import contextmanager
import gzip
from datetime import datetime
import logging


class Output:

    def __init__(self, name):
        self.name = name
        self._screen = config.args.screen
        self._pscreen = config.args.pscreen
        self._nscreen = config.args.nscreen
        self._logfile = config.args.log
        if self._logfile is None:
            logging.basicConfig(level=logging.DEBUG)
        else:
            logging.basicConfig(level=logging.DEBUG, filename=self._logfile)
        self.logger = logging.getLogger(__name__)
        pt.set_logger(self.logger)

    def screen(self, *args, **kw):
        if self._pscreen:
            print(*args, flush=True)
        elif self._nscreen:
            pt.sub_print(*args, **kw)
        elif self._screen:
            pt.single_print(*args, **kw)
        else:
            pass

    @pt.rank_zero
    def info(self, msg):
        self.logger.info(msg)

    @pt.rank_zero
    def warning(self, msg):
        self.logger.warning(msg)

    @staticmethod
    def exception(err):
        pt.exception(err)
        # if '{}'.format(err) == 'MPI_ERR_INTERN: internal error':
        #     # Known Issues: Allocating too much memory
        #     self.screen("Attempting to handle MPI error gracefully.\nAborting MPI...")
        #     pt.abort()

    def output(self, *args):
        pass


@contextmanager
def optional_write(file, mode, *args, openfn=None, **kwargs):
    # Note: this is suboptimal in that whatever write operations
    # are still performed ; this is negligible compared to the computation, at the moment.
    """If file is None, yields a dummy file object."""
    if file is None:
        return
    else:
        if openfn is None:
            openfn = gzip.open if file.endswith('.gz') else open
        with openfn(file, mode, *args, **kwargs) as open_file:
            with print_doing(f'Writing file "{file}"'):
                yield open_file


@contextmanager
def print_doing(msg, sep='', flush=True, end='', **kwargs):
    """Let the user know that the code is performing MSG. Also makes code more self-documenting."""
    start_time = datetime.now()
    print(msg, '...', sep=sep, flush=flush, end=end, **kwargs)
    yield
    print(f"Done! ({datetime.now()-start_time})")



from fitsnap3.calculators.calculator import Calculator
from fitsnap3.calculators.template_calculator import Template
from fitsnap3.calculators.lammps_snap import LammpsSnap


def calculator(calculator_name):
    """Calculator Factory"""
    instance = search(calculator_name)
    instance.__init__(calculator_name)
    return instance


def search(calculator_name):
    instance = None
    for cls in Calculator.__subclasses__():
        if cls.__name__.lower() == calculator_name.lower():
            instance = Calculator.__new__(cls)

    if instance is None:
        raise IndexError("{} was not found in fitsnap calculators".format(calculator_name))
    else:
        return instance



# <!----------------BEGIN-HEADER------------------------------------>
# ## FitSNAP3
# A Python Package For Training SNAP Interatomic Potentials for use in the LAMMPS molecular dynamics package
#
# _Copyright (2016) Sandia Corporation.
# Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
# the U.S. Government retains certain rights in this software.
# This software is distributed under the GNU General Public License_
# ##
#
# #### Original author:
#     Aidan P. Thompson, athomps (at) sandia (dot) gov (Sandia National Labs)
#     http://www.cs.sandia.gov/~athomps
#
# #### Key contributors (alphabetical):
#     Mary Alice Cusentino (Sandia National Labs)
#     Nicholas Lubbers (Los Alamos National Lab)
#     Charles Sievers (Sandia National Labs)
#     Adam Stephens (Sandia National Labs)
#     Mitchell Wood (Sandia National Labs)
#
# #### Additional authors (alphabetical):
#     Elizabeth Decolvenaere (D. E. Shaw Research)
#     Stan Moore (Sandia National Labs)
#     Steve Plimpton (Sandia National Labs)
#     Gary Saavedra (Sandia National Labs)
#     Peter Schultz (Sandia National Labs)
#     Laura Swiler (Sandia National Labs)
#
# <!-----------------END-HEADER------------------------------------->

from fitsnap3.parallel_tools import pt
from fitsnap3.scrapers.scraper_factory import scraper
from fitsnap3.calculators.calculator_factory import calculator
from fitsnap3.solvers.solver_factory import solver
from fitsnap3.io.output import output
from fitsnap3.io.input import config


class FitSnap:
    def __init__(self):
        self.scraper = scraper(config.sections["SCRAPER"].scraper)
        self.data = []
        self.calculator = calculator(config.sections["CALCULATOR"].calculator)
        self.solver = solver(config.sections["SOLVER"].solver)
        self.fit = None

    @pt.single_timeit
    def scrape_configs(self):
        self.scraper.scrape_groups()
        self.scraper.divvy_up_configs()
        self.data = self.scraper.scrape_configs()
        del self.scraper

    @pt.single_timeit
    def process_configs(self):
        self.calculator.create_a()
        for i, configuration in enumerate(self.data):
            self.calculator.process_configs(configuration, i)

    @pt.sub_rank_zero
    @pt.single_timeit
    def perform_fit(self):
        self.solver.perform_fit()
        self.solver.fit_gather()
        self.solver.error_analysis(self.data)

    @pt.single_timeit
    def write_output(self):
        output.output(self.solver.fit, self.solver.errors)



"""Template Class"""
from fitsnap3.io.sections.sections import Section


class Default(Section):

    def __init__(self, name, config, args):
        # let parent hold config and args
        super().__init__(name, config, args)
        # run init methods to populate class
        # delete config and args to rely only on child's members
        self.delete()
